# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import abstractmethod

from . import base_matcher
from . import edge_matcher
from . import function
from . import node_matcher
from . import walk_matcher


class BaseGraphFilter(object):
    """
    Base class to implement graph filtering by nodes, edges and sequences of nodes.
    """

    def filter(self, matcher: base_matcher.BaseMatcher) -> list:
        """
        Receive a matcher and return a list of matches in the graph.

        Args:
            matcher: Object of type BaseMatcher.

        Returns:
            List of matches.
        """

        # Return the nodes that matches the matcher object.
        if function.is_node_matcher(matcher):
            return self._node_filter(matcher)

        # Return the edges that matches the matcher object.
        elif function.is_edge_matcher(matcher):
            return self._edge_filter(matcher)

        # Return a list of nodes that match the matcher.
        elif function.is_walk_matcher(matcher):
            return self._walk_filter(matcher)
        else:
            raise NotImplemented  # pragma: no cover

    @abstractmethod
    def _node_filter(self, node_matcher: node_matcher.BaseNodeMatcher) -> list:
        """
        Returns the nodes in the graph that matches the matcher object.

        Args:
            node_matcher: Matcher object to apply on nodes in the graph.

        Returns:
            List of nodes that match the node_matcher.
        """
        pass  # pragma: no cover

    @abstractmethod
    def _edge_filter(self, edge_matcher: edge_matcher.BaseEdgeMatcher) -> list:
        """
        Returns the edges in the graph that match the matcher object.

        Args:
            edge_matcher: Matcher object to apply on the edges.

        Returns:
            List of edges that match.
        """
        pass  # pragma: no cover

    @abstractmethod
    def _walk_filter(self, walk_matcher: walk_matcher.WalkMatcherList) -> list:
        """
        Search for a list of nodes which match the list in walk_matcher. and return it.
        If one the nodes in the list (that was found in the graph) has more than one output,
        this list is not returned.

        Args:
            walk_matcher: WalkMatcherList with a list of nodes to match.

        Returns:
            A list of nodes which match the list in walk_matcher.
        """
        pass  # pragma: no cover



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Callable

from model_compression_toolkit import CoreConfig
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import TENSORFLOW, FOUND_TF
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi import KPI
from model_compression_toolkit.core.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfigV2
from model_compression_toolkit.core.common.target_platform.targetplatform2framework import TargetPlatformCapabilities
from model_compression_toolkit.core.runner import core_runner, _init_tensorboard_writer
from model_compression_toolkit.ptq.runner import ptq_runner

if FOUND_TF:
    import tensorflow as tf
    import tensorflow_model_optimization as tfmot
    from model_compression_toolkit.qat.keras.quantizer.config_factory import WeightQuantizeConfig

    from model_compression_toolkit.core.keras.default_framework_info import DEFAULT_KERAS_INFO
    from model_compression_toolkit.core.keras.keras_implementation import KerasImplementation
    from model_compression_toolkit.core.keras.keras_model_validation import KerasModelValidation
    from tensorflow.keras.models import Model
    from model_compression_toolkit.core.keras.constants import DEFAULT_TP_MODEL
    from model_compression_toolkit.qat.keras.quantizer.config_factory import QUANTIZATION_CONFIGS_DICT
    from model_compression_toolkit.qat.keras.qat_model_builder import QATKerasModelBuilder

    from model_compression_toolkit import get_target_platform_capabilities
    DEFAULT_KERAS_TPC = get_target_platform_capabilities(TENSORFLOW, DEFAULT_TP_MODEL)


    def keras_quantization_aware_training_init(in_model: Model,
                                               representative_data_gen: Callable,
                                               target_kpi: KPI = None,
                                               core_config: CoreConfig = CoreConfig(),
                                               fw_info: FrameworkInfo = DEFAULT_KERAS_INFO,
                                               target_platform_capabilities: TargetPlatformCapabilities = DEFAULT_KERAS_TPC):
        """
         Prepare a trained Keras model for quantization aware training. First the model quantization is optimized
         with post-training quantization, then the model layers are wrapped with QuantizeWrappers. The model is
         quantized using a symmetric quantization thresholds (power of two).
         The model is first optimized using several transformations (e.g. BatchNormalization folding to
         preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
         being collected for each layer's output (and input, depends on the quantization configuration).
         For each possible bit width (per layer) a threshold is then being calculated using the collected
         statistics. Then, if given a mixed precision config in the core_config, using an ILP solver we find
         a mixed-precision configuration, and set a bit-width for each layer. The model is built with fake_quant
         nodes for quantizing activation. Weights are kept as float and are quantized online while training by the
         quantization wrapper's weight quantizer.
         In order to limit the maximal model's size, a target KPI need to be passed after weights_memory
         is set (in bytes).

         Args:
             in_model (Model): Keras model to quantize.
             representative_data_gen (Callable): Dataset used for initial calibration.
             target_kpi (KPI): KPI object to limit the search of the mixed-precision configuration as desired.
             core_config (CoreConfig): Configuration object containing parameters of how the model should be quantized, including mixed precision parameters.
             fw_info (FrameworkInfo): Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.).  `Default Keras info <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/core/keras/default_framework_info.py>`_
             target_platform_capabilities (TargetPlatformCapabilities): TargetPlatformCapabilities to optimize the Keras model according to.

         Returns:

             A quantized model.
             User information that may be needed to handle the quantized model.
             Custom-Objects dictionary for loading the saved kers model.

         Examples:

             Import MCT:

             >>> import model_compression_toolkit as mct

             Import a Keras model:

             >>> from tensorflow.keras.applications.mobilenet_v2 import MobileNetV2
             >>> model = MobileNetV2()

             Create a random dataset generator:

             >>> import numpy as np
             >>> def repr_datagen(): return [np.random.random((1,224,224,3))]

             Create a MCT core config, containing the quantization configuration:

             >>> config = mct.CoreConfig()

             If mixed precision is desired, create a MCT core config with a mixed-precision configuration, to quantize a model with different bitwidths for different layers.
             The candidates bitwidth for quantization should be defined in the target platform model:

             >>> config = mct.CoreConfig(mixed_precision_config=MixedPrecisionQuantizationConfigV2())

             For mixed-precision set a target KPI object:
             Create a KPI object to limit our returned model's size. Note that this value affects only coefficients
             that should be quantized (for example, the kernel of Conv2D in Keras will be affected by this value,
             while the bias will not):

             >>> kpi = mct.KPI(model.count_params() * 0.75)  # About 0.75 of the model size when quantized with 8 bits.

             Pass the model, the representative dataset generator, the configuration and the target KPI to get a
             quantized model:

             >>> quantized_model, quantization_info, custom_objects = mct.keras_quantization_aware_training_init(model, repr_datagen, kpi, core_config=config)

             Use the quantized model for fine-tuning. For loading the model from file, use the custom_objects dictionary:

             >>> quantized_model = tf.keras.models.load_model(model_file, custom_objects=custom_objects)

             For more configuration options, please take a look at our `API documentation <https://sony.github.io/model_optimization/api/api_docs/modules/mixed_precision_quantization_config.html>`_.

         """
        KerasModelValidation(model=in_model,
                             fw_info=fw_info).validate()

        if core_config.mixed_precision_enable:
            if not isinstance(core_config.mixed_precision_config, MixedPrecisionQuantizationConfigV2):
                common.Logger.error("Given quantization config to mixed-precision facade is not of type "
                                    "MixedPrecisionQuantizationConfigV2. Please use keras_post_training_quantization API,"
                                    "or pass a valid mixed precision configuration.")

            common.Logger.info("Using experimental mixed-precision quantization. "
                               "If you encounter an issue please file a bug.")

        tb_w = _init_tensorboard_writer(fw_info)

        fw_impl = KerasImplementation()

        tg, bit_widths_config = core_runner(in_model=in_model,
                                            representative_data_gen=representative_data_gen,
                                            core_config=core_config,
                                            fw_info=fw_info,
                                            fw_impl=fw_impl,
                                            tpc=target_platform_capabilities,
                                            target_kpi=target_kpi,
                                            tb_w=tb_w)

        tg = ptq_runner(tg, fw_info, fw_impl, tb_w)

        qat_model, user_info = QATKerasModelBuilder(graph=tg, fw_info=fw_info).build_model()

        user_info.mixed_precision_cfg = bit_widths_config

        return qat_model, user_info, QUANTIZATION_CONFIGS_DICT


    def keras_quantization_aware_training_finalize(in_model: Model):
        """
         Convert a model fine-tuned by the user to a network without QuantizeWrappers. The exported
         model contains float (fake-quantized) parameters and fake-quantiztion layers for quantizing
         the activations

         Args:
             in_model (Model): Keras model to remove QuantizeWrappers.

         Returns:
             A quantized model without QuantizeWrappers.

         Examples:

             Import MCT:

             >>> import model_compression_toolkit as mct

             Import a Keras model:

             >>> from tensorflow.keras.applications.mobilenet_v2 import MobileNetV2
             >>> model = MobileNetV2()

             Create a random dataset generator:

             >>> import numpy as np
             >>> def repr_datagen(): return [np.random.random((1,224,224,3))]

             Create a MCT core config, containing the quantization configuration:

             >>> config = mct.CoreConfig()

             If mixed precision is desired, create a MCT core config with a mixed-precision configuration, to quantize a model with different bitwidths for different layers.
             The candidates bitwidth for quantization should be defined in the target platform model:

             >>> config = mct.CoreConfig(mixed_precision_config=MixedPrecisionQuantizationConfigV2())

             For mixed-precision set a target KPI object:
             Create a KPI object to limit our returned model's size. Note that this value affects only coefficients
             that should be quantized (for example, the kernel of Conv2D in Keras will be affected by this value,
             while the bias will not):

             >>> kpi = mct.KPI(model.count_params() * 0.75)  # About 0.75 of the model size when quantized with 8 bits.

             Pass the model, the representative dataset generator, the configuration and the target KPI to get a
             quantized model:

             >>> quantized_model, quantization_info, custom_objects = mct.keras_quantization_aware_training_init(model, repr_datagen, kpi, core_config=config)

             Use the quantized model for fine-tuning. For loading the model from file, use the custom_objects dictionary:

             >>> quantized_model = tf.keras.models.load_model(model_file, custom_objects=custom_objects)
             >>> quantized_model = mct.keras_quantization_aware_training_finalize(quantized_model)

         """

        def _export(layer):
            if isinstance(layer, tfmot.quantization.keras.QuantizeWrapper):
                if not isinstance(layer.quantize_config, tuple(QUANTIZATION_CONFIGS_DICT.values())):
                    Logger.error(f'Only supported quantization configs are {tuple(QUANTIZATION_CONFIGS_DICT.keys())}')
                if isinstance(layer.quantize_config, WeightQuantizeConfig):
                    new_layer = layer.layer.__class__.from_config(layer.layer.get_config())
                    with tf.name_scope(new_layer.name):
                        new_layer.build(layer.input_shape)
                    weights_list = []
                    for w in new_layer.weights:
                        val = None
                        for qw in layer.weights:
                            if w.name in qw.name:
                                if w.name.split('/')[-1].split(':')[0] in layer.quantize_config.weight_attrs:
                                    val = layer.quantize_config.get_weights_and_quantizers(layer.layer)[0][1](qw,
                                                                                                              False,
                                                                                                              layer.quantize_config.weight_quantizer.quantizer_parameters)
                                else:
                                    val = qw
                                val = val.numpy()
                        if val is None:
                            Logger.error(f'Could not match weight name: {w.name}')
                        weights_list.append(val)
                    new_layer.set_weights(weights_list)
                    new_layer.trainable = False
                    return new_layer
                else:
                    Logger.error(f'Undefined quantize_config')
            else:
                return layer

        # clone each layer in the model and apply _export to layers wrapped with a QuantizeWrapper.
        exported_model = tf.keras.models.clone_model(in_model, input_tensors=None, clone_function=_export)

        return exported_model

else:
    # If tensorflow or tensorflow_model_optimization are not installed,
    # we raise an exception when trying to use these functions.
    def keras_quantization_aware_training_init(*args, **kwargs):
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using keras_quantization_aware_training_init. '
                        'Could not find Tensorflow package.')

    def keras_quantization_aware_training_finalize(*args, **kwargs):
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using keras_quantization_aware_training_finalize. '
                        'Could not find Tensorflow package.')



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import copy
import numpy as np
from typing import Tuple, Callable
from model_compression_toolkit.core.common.logger import Logger
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.graph.base_graph import Graph
from model_compression_toolkit.core.common.graph.graph_matchers import EdgeMatcher, NodeOperationMatcher
from model_compression_toolkit.core.common.graph.base_node import BaseNode


class Conv2DCollapsing(common.BaseSubstitution):
    """
    Collapse Conv2D into preceding Conv2D (Not non-linear activation between them)
    """
    def __init__(self,
                 first_node: NodeOperationMatcher,
                 second_node: NodeOperationMatcher,
                 conv2d_collapsing_fn: Callable,
                 kernel_str: str,
                 kernel_size: str,
                 bias_str: str,
                 use_bias_str: str,
                 strides_str: str,
                 paddings_str: str,
                 dilations_str: str,
                 groups_str: str,
                 filters_str: str,
                 data_format_str: str = None,
                 layer_name_str: str = None):
        """
        Collapsing Conv2D node (first node) to Conv2D node (second node).
        Args:
            first_node: Node matcher for convolution type nodes.
            second_node: Node matcher for convolution type nodes.
            conv2d_collapsing_fn: Function for updating the convolution kernel and bias
            kernel_str: The framework specific attribute name of the convolution layer's weight/kernel.
            kernel_size_str: The framework specific attribute name of the convolution layer's kernel size.
            bias_str: The framework specific attribute name of the convolution layer's bias.
            use_bias_str: The framework specific attribute name of the convolution layer's bias flag.
            strides_str: The framework specific attribute name of the convolution layer's strides.
            paddings_str: The framework specific attribute name of the convolution layer's paddings.
            dilations_str: The framework specific attribute name of the convolution layer's dilations.
            groups_str: The framework specific attribute name of the convolution layer's groups.
            filters_str: The framework specific attribute name of the convolution layer's filters.
            data_format_str: The framework specific attribute name of the convolution layer's data format.
            layer_name_str: The framework specific attribute name of layer's name.
        """
        super().__init__(matcher_instance=EdgeMatcher(first_node, second_node))
        self.conv2d_collapsing_fn = conv2d_collapsing_fn
        self.kernel_str = kernel_str
        self.kernel_size = kernel_size
        self.bias_str = bias_str
        self.use_bias_str = use_bias_str
        self.strides_str = strides_str
        self.paddings_str = paddings_str
        self.dilations_str = dilations_str
        self.groups_str = groups_str
        self.filters_str = filters_str
        self.data_format_str = data_format_str
        self.layer_name_str = layer_name_str

    def substitute(self,
                   graph: Graph,
                   edge_nodes: Tuple[BaseNode, BaseNode]) -> Graph:
        """
        Collapse linear layer into preceding linear layers.
        Convolution condition:
        |----------------------------|      |--------------|
        | Conv2D k1xk2 -> Conv2D 1x1 | ---> | Conv2D k1xk2 |
        |----------------------------|      |--------------|
        Args:
            graph: Graph we apply the substitution on.
            edge_nodes: Tuple of two linear nodes
        Returns:
            Graph after applying the substitution.
        """

        first_node = edge_nodes[0]
        second_node = edge_nodes[1]

        # If the linear operator is part of a reused group (it is the "base" node, or a reused node),
        # we should skip the substitution.
        if first_node.reuse or first_node.reuse_group is not None:
            return graph
        if second_node.reuse or second_node.reuse_group is not None:
            return graph

        # If there is an extra connection between these two nodes skip the substitution
        if len(graph.get_next_nodes(first_node)) > 1 or len(graph.get_prev_nodes(second_node)) > 1:
            return graph

        # Skip if convolution's data format is 'channels_first'
        if self.data_format_str is not None:
            data_format1 = first_node.framework_attr[self.data_format_str]
            data_format2 = second_node.framework_attr[self.data_format_str]
            if data_format1 == 'channels_first' or data_format2 == 'channels_first':
                Logger.warning("No supported Conv2D with 'channels_first' in block collapsing!")
                return graph

        # Get kernels and strides
        kernel2_size = second_node.framework_attr[self.kernel_size]
        paddings2 = second_node.framework_attr[self.paddings_str]
        strides1 = first_node.framework_attr[self.strides_str]
        strides2 = second_node.framework_attr[self.strides_str]
        dilations1 = first_node.framework_attr[self.dilations_str]
        dilations2 = second_node.framework_attr[self.dilations_str]
        groups1 = first_node.framework_attr[self.groups_str]
        groups2 = second_node.framework_attr[self.groups_str]

        # Check if convolutions parameters satisfy with the collapsing conditions, if not skip the
        # Collapsing 2 linear convolutions can be done only when:
        # 1. No strides and dilations
        # 2. First kernel is (k1,k2), second kernel is (1,1)
        # 3. No convolution groups
        # 4. No padding between the convolutions (padding2):
        #   a. Keras: 1x1 is always zero padding for padding2='same' or 'valid'
        #   b. Pytorch: padding2 can be 'same' or 'valid', and also can be padding=(p1,p2). for the latter we support only padding=(0,0).
        skip = True
        if kernel2_size == (1,1) and strides1 == strides2 == dilations1 == dilations2 == (1,1) \
                and groups1 == groups2 == 1 and (isinstance(paddings2, str) or paddings2 == (0,0)):
            skip = False
        if skip:
            return graph

        # New collapsed weights
        kernel, bias = self.conv2d_collapsing_fn(first_node, second_node, self.kernel_str, self.bias_str)

        # If the number of parameters is bigger after the collapsing, skip the substitution
        kernel1 = first_node.get_weights_by_keys(self.kernel_str)
        kernel2 = second_node.get_weights_by_keys(self.kernel_str)
        num_parameters = np.prod(kernel1.shape) + np.prod(kernel2.shape)
        new_num_parameters = np.prod(kernel.shape)
        if num_parameters < new_num_parameters:
            return graph

        num_nodes_before_substition = len(graph.nodes)
        num_edges_before_substition = len(graph.edges)

        # New collapsed node
        conv_collapsed = copy.deepcopy(first_node)
        conv_collapsed_name = first_node.name + '_' + second_node.name + '_collapsed'
        conv_collapsed.name = conv_collapsed_name
        weights_dict = {self.kernel_str: kernel}
        conv_collapsed.framework_attr[self.use_bias_str] = False
        if bias is not None:
            weights_dict.update({self.bias_str: bias})
            conv_collapsed.framework_attr[self.use_bias_str] = True
        conv_collapsed.weights = weights_dict
        conv_collapsed.input_shape = first_node.input_shape
        conv_collapsed.framework_attr[self.filters_str] = second_node.framework_attr[self.filters_str]
        if self.layer_name_str is not None:
            conv_collapsed.framework_attr[self.layer_name_str] = conv_collapsed_name

        # Update graph
        graph.add_node(conv_collapsed)
        graph.reconnect_out_edges(current_node=second_node, new_node=conv_collapsed)
        graph.reconnect_in_edges(current_node=first_node, new_node=conv_collapsed)
        graph.replace_output_node(current_node=second_node, new_node=conv_collapsed)

        graph.remove_edge(first_node, second_node)
        graph.remove_node(first_node)
        graph.remove_node(second_node)

        # Sanity check
        assert num_nodes_before_substition - len(graph.nodes) == 1
        assert num_edges_before_substition - len(graph.edges) == 1

        return graph



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.constants import FOUND_TORCH
from model_compression_toolkit.core.common import Logger
from model_compression_toolkit.core.common.constants import PYTORCH
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig
from model_compression_toolkit.core.common.target_platform import TargetPlatformCapabilities
from model_compression_toolkit.core.common.mixed_precision.kpi_tools.kpi import KPI
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.core.runner import core_runner, _init_tensorboard_writer
from model_compression_toolkit.gptq.runner import gptq_runner
from model_compression_toolkit.core.exporter import export_model
from model_compression_toolkit.core.analyzer import analyzer_model_quantization
from model_compression_toolkit import CoreConfig
from model_compression_toolkit.core.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfigV2

LR_DEFAULT = 1e-4
LR_REST_DEFAULT = 1e-4
LR_BIAS_DEFAULT = 1e-4
LR_QUANTIZATION_PARAM_DEFAULT = 1e-4

if FOUND_TORCH:
    from model_compression_toolkit.core.pytorch.default_framework_info import DEFAULT_PYTORCH_INFO
    from model_compression_toolkit.core.pytorch.pytorch_implementation import PytorchImplementation
    from model_compression_toolkit.core.pytorch.constants import DEFAULT_TP_MODEL
    from model_compression_toolkit.gptq.pytorch.gptq_loss import multiple_tensors_mse_loss
    import torch
    from torch.nn import Module
    from torch.optim import Adam, Optimizer
    from model_compression_toolkit import get_target_platform_capabilities


    DEFAULT_PYTORCH_TPC = get_target_platform_capabilities(PYTORCH, DEFAULT_TP_MODEL)


    def get_pytorch_gptq_config(n_iter: int,
                                optimizer: Optimizer = Adam([torch.Tensor([])], lr=LR_DEFAULT),
                                optimizer_rest: Optimizer = Adam([torch.Tensor([])], lr=LR_REST_DEFAULT),
                                loss: Callable = multiple_tensors_mse_loss,
                                log_function: Callable = None) -> GradientPTQConfig:
        """
        Create a GradientPTQConfig instance for Pytorch models.

        args:
            n_iter (int): Number of iterations to fine-tune.
            optimizer (Optimizer): Pytorch optimizer to use for fine-tuning for auxiliry variable.
            optimizer_rest (Optimizer): Pytorch optimizer to use for fine-tuning of the bias variable.
            loss (Callable): loss to use during fine-tuning. should accept 4 lists of tensors. 1st list of quantized tensors, the 2nd list is the float tensors, the 3rd is a list of quantized weights and the 4th is a list of float weights.
            log_function (Callable): Function to log information about the gptq process.

        returns:
            a GradientPTQConfig object to use when fine-tuning the quantized model using gptq.

        Examples:

            Create a GradientPTQConfig to run for 5 iteration:

            >>> gptq_conf = get_pytorch_gptq_config(n_iter=5)

            Other Tensorflow optimizers can be passed with dummy params:

            >>> gptq_conf = get_pytorch_gptq_config(n_iter=3, optimizer=torch.optim.Adam([torch.Tensor(1)]))

            The configuration can be passed to :func:`~model_compression_toolkit.pytorch_post_training_quantization` in order to quantize a pytorch model using gptq.

        """
        bias_optimizer = Adam([torch.Tensor([])], lr=LR_BIAS_DEFAULT)
        optimizer_quantization_parameter = Adam([torch.Tensor([])], lr=LR_QUANTIZATION_PARAM_DEFAULT)
        return GradientPTQConfig(n_iter,
                                 optimizer,
                                 optimizer_rest=optimizer_rest,
                                 loss=loss,
                                 log_function=log_function,
                                 train_bias=True,
                                 optimizer_quantization_parameter=optimizer_quantization_parameter,
                                 optimizer_bias=bias_optimizer)


    def pytorch_gradient_post_training_quantization_experimental(model: Module,
                                                                 representative_data_gen: Callable,
                                                                 target_kpi: KPI = None,
                                                                 core_config: CoreConfig = CoreConfig(),
                                                                 fw_info: FrameworkInfo = DEFAULT_PYTORCH_INFO,
                                                                 gptq_config: GradientPTQConfig = None,
                                                                 target_platform_capabilities: TargetPlatformCapabilities = DEFAULT_PYTORCH_TPC):
        """
        Quantize a trained Pytorch module using post-training quantization.
        By default, the module is quantized using a symmetric constraint quantization thresholds
        (power of two) as defined in the default TargetPlatformCapabilities.
        The module is first optimized using several transformations (e.g. BatchNormalization folding to
        preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
        being collected for each layer's output (and input, depends on the quantization configuration).
        Thresholds are then being calculated using the collected statistics and the module is quantized
        (both coefficients and activations by default).
        If gptq_config is passed, the quantized weights are optimized using gradient based post
        training quantization by comparing points between the float and quantized modules, and minimizing the
        observed loss.
        Then, the quantized weights are optimized using gradient based post
        training quantization by comparing points between the float and quantized models, and minimizing the observed
        loss.

        Args:
            model (Module): Pytorch model to quantize.
            representative_data_gen (Callable): Dataset used for calibration.
            target_kpi (KPI): KPI object to limit the search of the mixed-precision configuration as desired.
            core_config (CoreConfig): Configuration object containing parameters of how the model should be quantized, including mixed precision parameters.
            fw_info (FrameworkInfo): Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.). `Default PyTorch info <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/core/pytorch/default_framework_info.py>`_
            gptq_config (GradientPTQConfig): Configuration for using gptq (e.g. optimizer).
            target_platform_capabilities (TargetPlatformCapabilities): TargetPlatformCapabilities to optimize the PyTorch model according to. `Default PyTorch TPC <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/core/tpc_models/pytorch_tp_models/pytorch_default.py>`_

        Returns:
            A quantized module and information the user may need to handle the quantized module.

        Examples:

            Import a Pytorch module:

            >>> import torchvision.models.mobilenet_v2 as models
            >>> module = models.mobilenet_v2()

            Create a random dataset generator:

            >>> import numpy as np
            >>> def repr_datagen(): return [np.random.random((1,224,224,3))]

            Import mct and pass the module with the representative dataset generator to get a quantized module:

            >>> import model_compression_toolkit as mct
            >>> quantized_module, quantization_info = mct.pytorch_gradient_post_training_quantization_experimental(module, repr_datagen)

        """

        if core_config.mixed_precision_enable:
            if not isinstance(core_config.mixed_precision_config, MixedPrecisionQuantizationConfigV2):
                common.Logger.error("Given quantization config to mixed-precision facade is not of type "
                                    "MixedPrecisionQuantizationConfigV2. Please use keras_post_training_quantization API,"
                                    "or pass a valid mixed precision configuration.")

            common.Logger.info("Using experimental mixed-precision quantization. "
                               "If you encounter an issue please file a bug.")

        tb_w = _init_tensorboard_writer(fw_info)

        fw_impl = PytorchImplementation()

        # ---------------------- #
        # Core Runner
        # ---------------------- #
        graph, bit_widths_config = core_runner(in_model=model,
                                               representative_data_gen=representative_data_gen,
                                               core_config=core_config,
                                               fw_info=fw_info,
                                               fw_impl=fw_impl,
                                               tpc=target_platform_capabilities,
                                               target_kpi=target_kpi,
                                               tb_w=tb_w)

        # ---------------------- #
        # GPTQ Runner
        # ---------------------- #
        graph_gptq = gptq_runner(graph, gptq_config, representative_data_gen, fw_info, fw_impl, tb_w)
        if core_config.debug_config.analyze_similarity:
            analyzer_model_quantization(representative_data_gen, tb_w, graph_gptq, fw_impl, fw_info)

        # ---------------------- #
        # Export
        # ---------------------- #
        quantized_model, user_info = export_model(graph_gptq, fw_info, fw_impl, tb_w, bit_widths_config)

        return quantized_model, user_info

else:
    # If torch is not installed,
    # we raise an exception when trying to use these functions.
    def get_pytorch_gptq_config(*args, **kwargs):
        Logger.critical('Installing Pytorch is mandatory '
                        'when using pytorch_gradient_post_training_quantization_experimental. '
                        'Could not find torch package.')


    def pytorch_gradient_post_training_quantization_experimental(*args, **kwargs):
        Logger.critical('Installing Pytorch is mandatory '
                        'when using pytorch_gradient_post_training_quantization_experimental. '
                        'Could not find the torch package.')



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Dict, Any, List

import numpy as np
import tensorflow as tf

from model_compression_toolkit import GumbelConfig
from model_compression_toolkit.core.keras.quantizer.base_quantizer import BaseTrainableQuantizer
from model_compression_toolkit.core.common.defaultdict import DefaultDict
from model_compression_toolkit.core import common
from model_compression_toolkit.gptq.keras.quantizer import kernel_functions
from model_compression_toolkit.gptq.keras.quantizer.gumbel_rounding.gumbel_softmax import sample_gumbel
from model_compression_toolkit.gptq.common import gptq_constants
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper
from tensorflow.python.framework.tensor_shape import TensorShape

P_INIT = 0.01


def _init_aux_var(w_shape: List[int], m: int, p: float = P_INIT) -> np.ndarray:
    """
    This function generate a random pi matrix for Gumbel Rounding
    Args:
        w_shape(List[int]): A list of integers that represent the shape of the weights tensor to be quantization
        p(float): A floating point number that represent the probability of non round options of pi matrix.
        m(int):  An integer that define the number of shift

    Returns: A numpy array of pi tensor

    """
    m_hat = m // 2
    shift = -np.log(-np.log(1 - p))
    n = np.random.randn(*[m, *w_shape]) * np.sqrt(np.power(np.pi, 2) / 6)
    n[m_hat, :] += shift
    return n


def _init_shift_var(m: int) -> List[int]:
    """
    This function generate an list of 2*m+1 from -m to m
    Args:
        m: An integer value the represent m

    Returns: A list of size m

    """
    m_hat = m // 2
    aux_index_shift = [-m_hat + i for i in range(m)]
    return aux_index_shift


class GumbelRoundingBase(BaseTrainableQuantizer):
    def __init__(self,
                 num_bits: int,
                 per_axis: bool,
                 signed: bool,
                 symmetric: bool,
                 power_of_two: bool,
                 quantization_parameter_learning: bool,
                 quantization_axis: int,
                 gumbel_config: GumbelConfig,
                 max_lsbs_change_map: dict = DefaultDict({}, lambda: 1),
                 max_iteration: int = 10000):
        """
        A base class for GumRounding

        Args:
            num_bits: Number of bits to use for the quantization.
            per_axis: Whether to quantize per-channel or per-tensor.
            signed: Signedness to use for the quantization range.
            symmetric:  Whether to quantize is symmetric.
            power_of_two: Whether to quantize is power-of-two.
            quantization_parameter_learning: A bool flag state if the quantizer parameter are trainable
            quantization_axis: Axis of tensor to use for the quantization.
            gumbel_config: A class with the gumbel rounding configurations.
            max_lsbs_change_map: a mapping between number of bits to max lsb change.
            max_iteration: The number of iteration of gptq.
        """
        self.num_bits = num_bits
        self.per_axis = per_axis
        self.signed = signed
        self.quantization_axis = quantization_axis
        self.max_iteration = max_iteration
        self.power_of_two = power_of_two
        self.symmetric = symmetric
        self.quantization_parameter_learning = quantization_parameter_learning
        self.temperature_learning = gumbel_config.temperature_learning
        self.quantizer_parameters = {}
        self.gumbel_config = gumbel_config

        self.max_lsbs_change_map = max_lsbs_change_map
        self.max_lsbs_change = max_lsbs_change_map.get(num_bits)
        self.m = 2 * self.max_lsbs_change + 1

        self.n_cycles = gumbel_config.n_cycles
        self.minimal_temp = gumbel_config.minimal_temp
        self.maximal_temp = gumbel_config.maximal_temp
        self.cycle_iterations = int(self.max_iteration / self.n_cycles)
        self.tau = None
        self.g_t = None
        self.p_t = None
        scale = self.cycle_iterations / (-2 * np.log(0.001))

        def tau_function(i):
            """
            A function the generate the gumbel temperature.
            Args:
                i: An int the represent the current iteration number

            Returns: A temperature value.

            """
            if i < (self.cycle_iterations - 1):
                index = ((i + 1) % self.cycle_iterations) / scale
            else:
                index = (i % self.cycle_iterations) / scale

            x = tf.exp(-index)
            return self.minimal_temp + (self.maximal_temp - self.minimal_temp) * x

        self.tau_function = tau_function
        self.w_shape = None
        self.update_gumbel_param = True

    def enable_update(self):
        self.update_gumbel_param = True

    def disable_update(self):
        self.update_gumbel_param = False

    def build(self, tensor_shape: TensorShape,
              name: str,
              layer: QuantizeWrapper) -> Dict[str, tf.Variable]:
        """
        Add min and max variables to layer.
        Args:
            tensor_shape: Tensor shape the quantizer quantize.
            name: Prefix of variables names.
            layer: Layer to add the variables to. The variables are saved
            in the layer's scope.

        Returns:
            Dictionary of new variables.
        """
        w_shape = kernel_functions.get_kernel(layer.weights).shape
        self.w_shape = w_shape

        ar_iter = layer.add_weight(
            name + gptq_constants.GPTQ_ITER,
            shape=(),
            initializer=tf.keras.initializers.Constant(0.0),
            trainable=False)

        auxvar_tensor = layer.add_weight(
            name + gptq_constants.AUXVAR,
            shape=[self.m, *self.w_shape],
            initializer=tf.keras.initializers.Constant(0.0),
            trainable=True)
        auxvar_tensor.assign(_init_aux_var(self.w_shape, self.m))

        temp_tensor = layer.add_weight(
            name + gptq_constants.TEMP,
            shape=[1, *self.w_shape],
            initializer=tf.keras.initializers.Constant(self.maximal_temp),
            trainable=True)

        shift_tensor = layer.add_weight(name + gptq_constants.AUXSHIFT,
                                        shape=self.m,
                                        initializer=tf.keras.initializers.Constant(0.0),
                                        trainable=False)
        shift_tensor.assign(_init_shift_var(self.m))

        self.quantizer_parameters = {gptq_constants.AUXVAR: auxvar_tensor,
                                     gptq_constants.GPTQ_ITER: ar_iter,
                                     gptq_constants.AUXSHIFT: shift_tensor,
                                     gptq_constants.TEMP: temp_tensor}
        return self.quantizer_parameters

    def get_aux_variable(self) -> tf.Tensor:
        return self.quantizer_parameters[gptq_constants.AUXVAR]

    def get_trainable_parameters(self) -> List[tf.Tensor]:
        """
        A function to get a list trainable of trainable parameters of the quantizer for GPTQ retraining

        Returns:
            A list of trainable Tensors

        """
        return [t for t in self.quantizer_parameters.values() if t.trainable]

    def __eq__(self, other: Any) -> bool:
        """
        Check if equals to another object.
        Args:
            other: Other object to compare.

        Returns:
            Whether they are equal or not.
        """
        if not isinstance(other, GumbelRoundingBase):
            return False

        return (self.num_bits == other.num_bits and
                self.per_axis == other.per_axis and
                self.symmetric == other.symmetric)

    def __ne__(self, other: Any) -> bool:
        """
        Check if not equals to another object.
        Args:
            other: Other object to compare.

        Returns:
            Whether they are differ or not.
        """
        return not self.__eq__(other)

    def get_config(self) -> Dict[str, Any]:
        """
        Returns: Configuration of TrainableQuantizer.
        """

        return {
            'num_bits': self.num_bits,
            'per_axis': self.per_axis,
            'symmetric': self.symmetric,
            'power_of_two': self.power_of_two
        }

    def update_iteration(self, training, ar_iter):
        if self.temperature_learning:
            self.tau = tf.minimum(tf.maximum(self.quantizer_parameters[gptq_constants.TEMP], self.minimal_temp),
                                  self.maximal_temp)
        else:
            self.tau = self.tau_function(ar_iter)
        if self.update_gumbel_param and training:

            if ar_iter % self.cycle_iterations == 0:
                self.quantizer_parameters[gptq_constants.TEMP].assign(
                    self.maximal_temp * np.ones(self.quantizer_parameters[gptq_constants.TEMP].shape))
            ar_iter.assign_add(1.0)
            self.g_t = sample_gumbel([self.m, *self.w_shape])

    def get_temperature_variable(self):
        return self.quantizer_parameters[gptq_constants.TEMP]

    def get_gumbel_probability(self):
        return self.p_t



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch.nn as nn
from typing import List
from model_compression_toolkit.gptq.pytorch.quantizer.quantizer_wrapper import WeightQuantizerWrapper


def get_trainable_parameters(fxp_model: nn.Module,
                             add_bias: bool = False,
                             quantization_parameters_learning: bool = False) -> (List[nn.Parameter], List[nn.Parameter], List[nn.Parameter]):
    """
    Get trainable parameters from all layers in a model

    Args:
        fxp_model: Model to get its trainable parameters.
        add_bias: Whether to include biases of the model (if there are) or not.
        quantization_parameters_learning: Whether to include quantization parameters of the model or not.
    Returns:
        A list of trainable variables in a model. Each item is a list of a layers weights.
    """

    trainable_aux_weights = nn.ParameterList()
    trainable_threshold = nn.ParameterList()
    trainable_bias = nn.ParameterList()

    for layer in fxp_model.modules():
        if isinstance(layer, WeightQuantizerWrapper):
            trainable_aux_weights.append(layer.weight_quantizer.get_aux_variable())
            if quantization_parameters_learning:
                trainable_threshold.append(layer.weight_quantizer.get_quantization_variable())
            if add_bias and layer.op.bias is not None:
                trainable_bias.append(layer.op.bias)

    return trainable_aux_weights, trainable_bias, trainable_threshold


def get_weights_for_loss(fxp_model: nn.Module) -> [List, List]:
    """
    Get all float and quantized kernels for the GPTQ loss

    Args:
        fxp_model: Model to get its float and quantized weights.

    Returns:
        A list of float kernels, each item is the float kernel of the layer
        A list of quantized kernels, each item is the quantized kernel of the layer
    """

    flp_weights_list, fxp_weights_list = [], []
    for layer in fxp_model.modules():
        if isinstance(layer, WeightQuantizerWrapper):
            # Collect pairs of float and quantized weights per layer
            weights = layer.op.weight
            flp_weights_list.append(weights)
            fxp_weights_list.append(layer.weight_quantizer(weights))

    return flp_weights_list, fxp_weights_list



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List, Tuple

import model_compression_toolkit as mct
from model_compression_toolkit.core.common.target_platform import OpQuantizationConfig, TargetPlatformModel
from model_compression_toolkit.core.common.target_platform.op_quantization_config import QuantizationMethod

tp = mct.target_platform


def get_tp_model() -> TargetPlatformModel:
    """
    A method that generates a default target platform model, with base 8-bit quantization configuration and 8, 4, 2
    bits configuration list for mixed-precision quantization.
    NOTE: in order to generate a target platform model with different configurations but with the same Operators Sets
    (for tests, experiments, etc.), use this method implementation as a test-case, i.e., override the
    'get_op_quantization_configs' method and use its output to call 'generate_tp_model' with your configurations.

    Returns: A TargetPlatformModel object.

    """
    base_config, mixed_precision_cfg_list = get_op_quantization_configs()
    return generate_tp_model(default_config=base_config,
                             base_config=base_config,
                             mixed_precision_cfg_list=mixed_precision_cfg_list,
                             name='tflite_tp_model')


def get_op_quantization_configs() -> Tuple[OpQuantizationConfig, List[OpQuantizationConfig]]:
    """
    Creates a default configuration object for 8-bit quantization, to be used to set a default TargetPlatformModel.
    In addition, creates a default configuration objects list (with 8, 4 and 2 bit quantization) to be used as
    default configuration for mixed-precision quantization.

    Returns: An OpQuantizationConfig config object and a list of OpQuantizationConfig objects.

    """
    # Create a quantization config.
    # A quantization configuration defines how an operator
    # should be quantized on the modeled hardware:
    eight_bits = tp.OpQuantizationConfig(
        activation_quantization_method=QuantizationMethod.UNIFORM,
        weights_quantization_method=QuantizationMethod.SYMMETRIC,
        activation_n_bits=8,
        weights_n_bits=8,
        weights_per_channel_threshold=True,
        enable_weights_quantization=True,
        enable_activation_quantization=True,
        quantization_preserving=False,
        fixed_scale=None,
        fixed_zero_point=None,
        weights_multiplier_nbits=None
    )

    mixed_precision_cfg_list = [] # No mixed precision

    return eight_bits, mixed_precision_cfg_list


def generate_tp_model(default_config: OpQuantizationConfig,
                      base_config: OpQuantizationConfig,
                      mixed_precision_cfg_list: List[OpQuantizationConfig],
                      name: str) -> TargetPlatformModel:
    """
    Generates TargetPlatformModel with default defined Operators Sets, based on the given base configuration and
    mixed-precision configurations options list.

    Args
        default_config: A default OpQuantizationConfig to set as the TP model default configuration.
        base_config: An OpQuantizationConfig to set as the TargetPlatformModel base configuration for mixed-precision purposes only.
        mixed_precision_cfg_list: A list of OpQuantizationConfig to be used as the TP model mixed-precision
            quantization configuration options.
        name: The name of the TargetPlatformModel.

    Returns: A TargetPlatformModel object.

    """
    # Create a QuantizationConfigOptions, which defines a set
    # of possible configurations to consider when quantizing a set of operations (in mixed-precision, for example).
    # If the QuantizationConfigOptions contains only one configuration,
    # this configuration will be used for the operation quantization:
    default_configuration_options = tp.QuantizationConfigOptions([default_config])

    # Create a TargetPlatformModel and set its default quantization config.
    # This default configuration will be used for all operations
    # unless specified otherwise (see OperatorsSet, for example):
    generated_tpc = tp.TargetPlatformModel(default_configuration_options, name=name)

    # To start defining the model's components (such as operator sets, and fusing patterns),
    # use 'with' the TargetPlatformModel instance, and create them as below:
    with generated_tpc:
        # In TFLite, the quantized operator specifications constraint operators quantization
        # differently. For more details:
        # https://www.tensorflow.org/lite/performance/quantization_spec#int8_quantized_operator_specifications
        tp.OperatorsSet("NoQuantization",
                         tp.get_default_quantization_config_options().clone_and_edit(
                                  quantization_preserving=True))

        fc = tp.OperatorsSet("FullyConnected",
                              tp.get_default_quantization_config_options().clone_and_edit(
                                       weights_per_channel_threshold=False))

        tp.OperatorsSet("L2Normalization",
                         tp.get_default_quantization_config_options().clone_and_edit(
                                  fixed_zero_point=0, fixed_scale=1 / 128))
        tp.OperatorsSet("LogSoftmax",
                         tp.get_default_quantization_config_options().clone_and_edit(
                                  fixed_zero_point=127, fixed_scale=16 / 256))
        tp.OperatorsSet("Tanh",
                         tp.get_default_quantization_config_options().clone_and_edit(
                                  fixed_zero_point=0, fixed_scale=1 / 128))
        tp.OperatorsSet("Softmax",
                         tp.get_default_quantization_config_options().clone_and_edit(
                                  fixed_zero_point=-128, fixed_scale=1 / 256))
        tp.OperatorsSet("Logistic",
                         tp.get_default_quantization_config_options().clone_and_edit(
                                  fixed_zero_point=-128, fixed_scale=1 / 256))

        conv2d = tp.OperatorsSet("Conv2d")
        kernel = tp.OperatorSetConcat(conv2d, fc)

        relu = tp.OperatorsSet("Relu")
        elu = tp.OperatorsSet("Elu")
        activations_to_fuse = tp.OperatorSetConcat(relu, elu)

        batch_norm = tp.OperatorsSet("BatchNorm")
        bias_add = tp.OperatorsSet("BiasAdd")
        add = tp.OperatorsSet("Add")
        squeeze = tp.OperatorsSet("Squeeze",
                                   qc_options=tp.get_default_quantization_config_options().clone_and_edit(quantization_preserving=True))
        # ------------------- #
        # Fusions
        # ------------------- #
        # Source: https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/grappler/optimizers/remapper
        tp.Fusing([kernel, bias_add])
        tp.Fusing([kernel, bias_add, activations_to_fuse])
        tp.Fusing([conv2d, batch_norm, activations_to_fuse])
        tp.Fusing([conv2d, squeeze, activations_to_fuse])
        tp.Fusing([batch_norm, activations_to_fuse])
        tp.Fusing([batch_norm, add, activations_to_fuse])

    return generated_tpc



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any

import tensorflow as tf

if tf.__version__ < "2.6":
    from tensorflow.python.keras.layers.core import TFOpLambda, SlicingOpLambda
    from tensorflow.python.keras.engine.keras_tensor import KerasTensor
    from tensorflow.python.keras.engine.node import Node as KerasNode
else:
    from keras.layers.core import TFOpLambda, SlicingOpLambda
    from keras.engine.keras_tensor import KerasTensor
    from keras.engine.node import Node as KerasNode

from model_compression_toolkit.core.common.graph.base_node import BaseNode
from model_compression_toolkit.core.common.graph.functional_node import FunctionalNode

keras = tf.keras
layers = keras.layers

REUSED_IDENTIFIER = '_reused_'


def build_node(node: KerasNode,
               node_name_to_node: dict) -> BaseNode:
    """
    Build a node from a Keras node. A node contains all information to reconstruct the layer it's representing
    in a model:
    operation, layer configuration, path for instantiating the Keras layer the node has, weights, group of other
    nodes if it's a reused layer,
    input/output shape.
    Args:
        node: Node in the graph of a Keras model.
        node_name_to_node: Dictionary of already created nodes aims to identify reused layers.

    Returns:
        Graph node that was built from the Keras node.
    """
    keras_layer = node.layer  # get the layer the node represents.
    layer_config = keras_layer.get_config()  # layer configuration to reconstruct it.
    op_call_args = node.call_args
    op_call_kwargs = node.call_kwargs
    layer_class = type(keras_layer)  # class path to instantiating it in back2framework.
    weights = {v.name: v.numpy() for v in keras_layer.weights}  # layer's weights

    # If it's a node representing a reused layer, several nodes will contain the same layer instance.
    # Thus, the first time a node of a reused layer is being created, it's being build as a node of a non-reused layer,
    # while other nodes of this layer will be created as nodes of a reused layer with the suffix "_reused_i"
    # where i is the input/output index of the layer.
    is_reused = keras_layer.name in node_name_to_node
    if is_reused:
        # Mark the "base" node with its reused group.
        node_name_to_node[keras_layer.name].reuse_group = keras_layer.name
        io_index = 1
        while keras_layer.name + REUSED_IDENTIFIER + str(io_index) in node_name_to_node:  # find next unused io index
            io_index = io_index + 1
        reuse_group = keras_layer.name  # by the layer name we can gather nodes of this reused layer
        node_name = keras_layer.name + REUSED_IDENTIFIER + str(io_index)
    else:
        io_index = 0  # for non reused layers input/output index is 0
        reuse_group = None
        node_name = keras_layer.name
    input_shape = keras_layer.get_input_shape_at(io_index)
    output_shape = keras_layer.get_output_shape_at(io_index)

    if layer_class in [TFOpLambda, SlicingOpLambda]:
        # Some functional ops (such as tf.concat) should receive the input tensors as a list
        # and some are not (such as tf.multiply), so each FunctionalNode holds
        # a flag to indicate that.
        inputs_as_list = __is_functional_inputs_a_list(op_call_args)
        # Do not hold the tensors that are in op_call_args as they are
        # not needed. Thus, if the first element in op_call_args is a list of
        # Keras tensors, remove it from op_call_args.
        op_call_args = op_call_args[int(inputs_as_list):]
        node = FunctionalNode(node_name,
                              layer_config,
                              input_shape,
                              output_shape,
                              weights,
                              layer_class,
                              [arg for arg in op_call_args if not isinstance(arg, KerasTensor)],  # Do not hold the tensors that are in op_call_args
                              {k: v for k, v in op_call_kwargs.items() if not isinstance(v, KerasTensor)}, # In TF2.5 tensors are in kwargs as well.
                              is_reused,
                              reuse_group,
                              functional_op=keras_layer.function,
                              inputs_as_list=inputs_as_list)
    else:
        node = BaseNode(node_name,
                        layer_config,
                        input_shape,
                        output_shape,
                        weights,
                        layer_class,
                        is_reused,
                        reuse_group)

    node_name_to_node[node_name] = node
    return node


def __is_functional_inputs_a_list(op_call_args: Any) -> bool:
    """
    Check whether the input tensors should be passed as a list
    or not.

    Args:
        op_call_args: Arguments list to check.

    Returns:
        Whether the input tensors should be passed as a list or not.
    """

    if len(op_call_args) > 0 and isinstance(op_call_args[0], list):
        inputs_as_list = True
        for arg in op_call_args[0]:
            inputs_as_list = inputs_as_list and isinstance(arg, KerasTensor)
        return inputs_as_list
    return False



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable, Any
import numpy as np

from model_compression_toolkit import FrameworkInfo, KPI, CoreConfig
from model_compression_toolkit.core.common import Graph
from model_compression_toolkit.core.common.constants import FLOAT_BITWIDTH
from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.core.common.graph.edge import EDGE_SINK_INDEX
from model_compression_toolkit.core.common.target_platform import TargetPlatformCapabilities
from model_compression_toolkit.core.runner import read_model_to_graph, get_finalized_graph
from model_compression_toolkit.core.common.logger import Logger


def compute_kpi_data(in_model: Any,
                     representative_data_gen: Callable,
                     core_config: CoreConfig,
                     tpc: TargetPlatformCapabilities,
                     fw_info: FrameworkInfo,
                     fw_impl: FrameworkImplementation) -> KPI:
    """
    Compute KPI information that can be relevant for defining target KPI for mixed precision search.
    Calculates maximal activation tensor, sum of weights' parameters and total (sum of both).

    Args:
        in_model:  Model to build graph from (the model that intended to be quantized).
        representative_data_gen: Dataset used for calibration.
        core_config: CoreConfig containing parameters of how the model should be quantized.
        tpc: TargetPlatformCapabilities object that models the inference target platform and
                                              the attached framework operator's information.
        fw_info: Information needed for quantization about the specific framework.
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.

    Returns: A KPI object with the results.

    """

    graph = read_model_to_graph(in_model,
                                representative_data_gen,
                                tpc,
                                fw_info,
                                fw_impl)

    transformed_graph = get_finalized_graph(graph,
                                            tpc,
                                            core_config.quantization_config,
                                            fw_info,
                                            tb_w=None,
                                            fw_impl=fw_impl,
                                            mixed_precision_enable=core_config.mixed_precision_enable)

    # Compute parameters sum
    weights_params = compute_configurable_weights_params(graph=transformed_graph, fw_info=fw_info)
    total_weights_params = 0 if len(weights_params) == 0 else sum(weights_params)

    # Compute max activation tensor
    activation_output_sizes = compute_activation_output_sizes(graph=transformed_graph)
    max_activation_tensor_size = 0 if len(activation_output_sizes) == 0 else max(activation_output_sizes)

    # Compute total kpi - parameters sum + max activation tensor
    total_size = total_weights_params + max_activation_tensor_size

    # Compute BOPS kpi - total count of bit-operations for all configurable layers with kernel
    bops_count = compute_total_bops(graph=transformed_graph, fw_info=fw_info, fw_impl=fw_impl)
    bops_count = np.inf if len(bops_count) == 0 else sum(bops_count)

    return KPI(weights_memory=total_weights_params,
               activation_memory=max_activation_tensor_size,
               total_memory=total_size,
               bops=bops_count)


def compute_configurable_weights_params(graph: Graph, fw_info: FrameworkInfo) -> np.ndarray:
    """
    Computes a vector with the respective weights' parameters size for each weight configurable node.

    Args:
        graph: Finalized Graph object.
        fw_info: FrameworkInfo object about the specific framework
            (e.g., attributes of different layers' weights to quantize).

    Returns: A vector of node's weights memory sizes.

    """

    weights_params = []
    # Go over all nodes that have configurable weights.
    for n in graph.get_sorted_weights_configurable_nodes():
        node_num_weights_params = 0
        for attr in fw_info.get_kernel_op_attributes(n.type):
            if attr is not None:
                node_num_weights_params += n.get_weights_by_keys(attr).flatten().shape[0]

        weights_params.append(node_num_weights_params)

    return np.array(weights_params)


def compute_activation_output_sizes(graph: Graph) -> np.ndarray:
    """
    Computes a vector with the respective output tensor size for each activation configurable node.

    Args:
        graph: Finalized Graph object.

    Returns: A vector of node's activation output size.

    """

    activation_outputs = []
    # Go over all nodes that have configurable activation.
    for n in graph.get_sorted_activation_configurable_nodes():
        node_output_size = n.get_total_output_params()
        activation_outputs.append(node_output_size)

    return np.array(activation_outputs)


def compute_total_bops(graph: Graph, fw_info: FrameworkInfo, fw_impl: FrameworkImplementation) -> np.ndarray:
    """
    Computes a vector with the respective Bit-operations count for each configurable node that includes MAC operations.
    The computation assumes that the graph is a representation of a float model, thus, BOPs computation uses 32-bit.

    Args:
        graph: Finalized Graph object.
        fw_info: FrameworkInfo object about the specific framework
            (e.g., attributes of different layers' weights to quantize).
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.

    Returns: A vector of nodes' Bit-operations count.

    """

    bops = []

    # Go over all configurable nodes that have kernels.
    for n in graph.get_topo_sorted_nodes():
        if n.has_weights_to_quantize(fw_info):
            # If node doesn't have weights then its MAC count is 0, and we shouldn't consider it in the BOPS count.
            incoming_edges = graph.incoming_edges(n, sort_by_attr=EDGE_SINK_INDEX)
            if len(incoming_edges) != 1:
                Logger.warning(f"Can't compute BOPS metric for node {n.name} with multiple inputs.")

            node_mac = fw_impl.get_node_mac_operations(n, fw_info)

            node_bops = (FLOAT_BITWIDTH ** 2) * node_mac
            bops.append(node_bops)

    return np.array(bops)



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Callable

from model_compression_toolkit.core import common
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig
from model_compression_toolkit.core.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.core.common import FrameworkInfo
from model_compression_toolkit.core.common.graph.base_graph import Graph
from model_compression_toolkit.gptq.common.gptq_training import gptq_training

from model_compression_toolkit.core.common.visualization.tensorboard_writer import TensorboardWriter
from model_compression_toolkit.core.common.bias_correction.apply_bias_correction_to_graph import \
    apply_bias_correction_to_graph


def _apply_gptq(gptq_config: GradientPTQConfig,
                representative_data_gen: Callable,
                tb_w: TensorboardWriter,
                tg: Graph,
                tg_bias: Graph,
                fw_info: FrameworkInfo,
                fw_impl: FrameworkImplementation) -> Graph:
    """
    Apply GPTQ to improve accuracy of quantized model.
    Build two models from a graph: A teacher network (float model) and a student network (quantized model).
    and use the dataset generator to pass images through the teacher and student networks to get intermediate
    layers outputs and maximize their similarity.

    Args:
        gptq_config: Configuration for using GPTQ (e.g. optimizer).
        representative_data_gen: Dataset used for calibration.
        tb_w: TensorBoardWriter object to log events.
        tg: Float Reference Graph.
        tg_bias: Graph of quantized model.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.).
        fw_impl: Framework implementation per framework
    Returns:

    """
    if gptq_config is not None and gptq_config.n_iter > 0:
        common.Logger.info("Using experimental Gradient Based PTQ: If you encounter an issue "
                           "please file a bug. To disable it, do not pass a gptq configuration.")

        tg_bias = gptq_training(tg,
                                tg_bias,
                                gptq_config,
                                representative_data_gen,
                                fw_impl,
                                fw_info)

        if tb_w is not None:
            tb_w.add_graph(tg_bias, 'after_gptq')
    return tg_bias


def gptq_runner(tg: Graph,
                gptq_config: GradientPTQConfig,
                representative_data_gen: Callable,
                fw_info: FrameworkInfo,
                fw_impl: FrameworkImplementation,
                tb_w: TensorboardWriter) -> Graph:
    """
    Quantize a graph that has final weights candidates quantization configurations.
    Before we quantize the graph weights, we apply GPTQ to get an improved graph.

    Args:
        tg: Graph to apply GPTQ and to quantize.
        gptq_config: GradientPTQConfig with parameters about the tuning process.
        representative_data_gen: Dataset used for calibration.
        fw_info: Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.)
        fw_impl: FrameworkImplementation object with a specific framework methods implementation.
        tb_w: A TensorBoardWriter object initialized with the logger dir path if it was set, or None otherwise.

    Returns:
        A graph after model weights GPTQ fine-tuning.

    """
    #############################################
    # Apply Bias Correction
    #############################################
    tg_bias = apply_bias_correction_to_graph(tg,
                                             fw_impl=fw_impl)
    if tb_w is not None:
        tb_w.add_graph(tg_bias, 'after_bias_correction')
    #############################################
    # Gradient Based Post Training Quantization
    #############################################
    tg_gptq = _apply_gptq(gptq_config,
                          representative_data_gen,
                          tb_w,
                          tg,
                          tg_bias,
                          fw_info,
                          fw_impl)

    return tg_gptq



# Copyright 2021 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import copy
from typing import List

from model_compression_toolkit.core.common.graph.base_graph import Graph
from model_compression_toolkit.core.common.graph.base_node import BaseNode


def merge_models_nodes(inner_model_node: BaseNode,
                       outer_graph: Graph,
                       inner_graph: Graph) -> List[BaseNode]:
    """
    Given two MultiDiGraphs (one of an outer model and the second of the inner model), merge their nodes into
    a single nodes list representing the nodes that should be in a single MultiDiGraph after unrolling the inner graph.
    This is done by removing the inner model input nodes and the model node in the outer graph.
    Args:
        outer_graph: MultiDiGraph of the outer model.
        inner_graph: MultiDiGraph of the inner model.
        inner_model_node: Node of the inner model in the graph of the outer model.

    Returns:
        List of nodes that should be in a single graph after unrolling the inner graph.
    """
    # merge internal nodes
    res_nodes = copy.copy(list(outer_graph.nodes))
    res_nodes.extend(inner_graph.nodes)
    for input_node in inner_graph.get_inputs():  # inputs of inner model will no longer be needed.
        res_nodes.remove(input_node)
    # Remove the inner model node since we unroll it, and it's not going to be a node in the final graph.
    res_nodes.remove(inner_model_node)
    return res_nodes



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
from typing import Tuple, List
from model_compression_toolkit.core.common.user_info import UserInformation
from model_compression_toolkit.core import common
from model_compression_toolkit.core.common.graph.base_graph import BaseNode
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig
from model_compression_toolkit.core.common.framework_info import FrameworkInfo
from model_compression_toolkit.core.pytorch.default_framework_info import DEFAULT_PYTORCH_INFO
from model_compression_toolkit.core.pytorch.back2framework.pytorch_model_builder import PyTorchModelBuilder, PytorchModel
from model_compression_toolkit.core.common.graph.functional_node import FunctionalNode
from model_compression_toolkit.gptq.pytorch.quantizer.quantizer_wrapper import quantizer_wrapper


class GPTQPytorchModel(PytorchModel):
    """
    Class for GPTQ PyTorch model.
    """

    def __init__(self,
                 graph: common.Graph,
                 gptq_config: GradientPTQConfig,
                 append2output=None,
                 return_float_outputs: bool = True):
        """
        Args:
            graph: Graph to build the model from.
            gptq_config: Configuration for GPTQ optimization.
            append2output: Nodes to append to model's output.
            return_float_outputs: Whether the model returns float tensors or not.
        """

        super().__init__(graph,
                         append2output,
                         DEFAULT_PYTORCH_INFO,
                         return_float_outputs)

        for node in graph.nodes():
            if not isinstance(node, FunctionalNode):
                self.add_module(node.name, quantizer_wrapper(node, gptq_config))

    def _quantize_node_activations(self,
                                   node: BaseNode,
                                   input_tensors: List[torch.Tensor]) -> List[torch.Tensor]:
        """
        Quantize node's activation given input tensors.

        Args:
            node: Node to quantize its outputs.
            input_tensors: Input tensors of the node.

        Returns:
            Output of the node.

        """
        return node.final_activation_quantization_cfg.quantize_node_output(input_tensors)


class GPTQPytorchModelBuilder(PyTorchModelBuilder):
    """
    Builder of GPTQ Pytorch models.
    """

    def __init__(self,
                 graph: common.Graph,
                 gptq_config: GradientPTQConfig,
                 append2output=None,
                 return_float_outputs: bool = True):
        """

        Args:
            graph: Graph to build the model from.
            gptq_config: Configuration for GPTQ optimization.
            append2output: Nodes to append to model's output.
            return_float_outputs: Whether the model returns float tensors or not.
        """
        super().__init__(graph,
                         append2output,
                         DEFAULT_PYTORCH_INFO,
                         return_float_outputs)
        self.gptq_config = gptq_config

    def build_model(self) -> Tuple[PytorchModel, UserInformation]:
        """
        Build a GPTQ PyTorch model and return it.
        Returns:
            GPTQ PyTorch model and user information.
        """
        return GPTQPytorchModel(self.graph,
                                self.gptq_config,
                                self.append2output,
                                self.return_float_outputs), self.graph.user_info



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
import torch.nn as nn
from model_compression_toolkit.core.common import BaseNode, Logger
from model_compression_toolkit.gptq.common.gptq_config import GradientPTQConfig, RoundingType
from model_compression_toolkit.gptq.pytorch.quantizer.gptq_quantizer import BaseWeightQuantizer
from model_compression_toolkit.gptq.pytorch.quantizer.ste_rounding.ste_weights_quantizer import STEWeightQuantizer
from model_compression_toolkit.core.pytorch.back2framework.instance_builder import node_builder
from model_compression_toolkit.core.pytorch.constants import KERNEL
from model_compression_toolkit.core.pytorch.utils import to_torch_tensor


class WeightQuantizerWrapper(nn.Module):

    def __init__(self, node: BaseNode, gptq_config: GradientPTQConfig, weight_quantizer: BaseWeightQuantizer):
        """
        Construct a Pytorch model that constitutes as a wrapper for a Pytorch layer, built from a given graph node.
        Args:
            node: Node to build its Pytorch quantizer wrapper.
            gptq_config: GradientPTQConfig object with parameters about the tuning process.
            weight_quantizer: BaseWeightQuantizer object for gradient based weight quantizer
        """
        super().__init__()

        # loading operation
        self.op = node.type(**node.framework_attr)

        # loading the weights from the graph node (weights of the trained model)
        self.op.load_state_dict({k: torch.Tensor(v) for k, v in node.weights.items()}, strict=False)
        self.float_weight = to_torch_tensor(getattr(self.op, KERNEL)).detach()

        # replace non-gradient needed nn.Parameter with gradient needed torch.tensor
        delattr(self.op, KERNEL)
        setattr(self.op, KERNEL, self.float_weight)
        setattr(getattr(self.op, KERNEL), 'requires_grad', True)

        # quantizer
        self.weight_quantizer = weight_quantizer(node.final_weights_quantization_cfg, gptq_config, self.float_weight.shape)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Weight fake quantizer wrapper
        Args:
            x: input to layer.
        Returns:
            Output of layer after using operation with fake quantized weights
        """
        # Run weight quantizer
        setattr(self.op, KERNEL, self.weight_quantizer(self.float_weight))
        # Do computation
        return self.op(x)


def quantizer_wrapper(node: BaseNode, gptq_config: GradientPTQConfig) -> nn.Module:
    """
    Construct a Pytorch model that constitutes as a wrapper for a Pytorch layer, built from a given graph node.
    Args:
        node: Node to build its Pytorch layer.
        gptq_config: GradientPTQConfig with parameters about the tuning process.
    """
    if node.is_weights_quantization_enabled():
        if gptq_config.rounding_type == RoundingType.STE:
            node_instance = WeightQuantizerWrapper(node, gptq_config, STEWeightQuantizer)
        else:
            Logger.critical('No support for GumbelRounding GPTQ yet. Work in progress..')
    else:
        # No quantization
        node_instance = node_builder(node)

    return node_instance



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
from typing import Any, List
from model_compression_toolkit.core.common.graph.base_graph import Graph
from model_compression_toolkit.core.common.graph.base_node import BaseNode
from model_compression_toolkit.core.common.target_platform.targetplatform2framework import TargetPlatformCapabilities
from model_compression_toolkit.core.common.target_platform.targetplatform2framework.layer_filter_params import LayerFilterParams


def filter_fusing_patterns(fusing_patterns: List[List[Any]], node: BaseNode, idx: int = 0) -> List[List[Any]]:
    """
    Update relevant fusing patterns object if layer number 'idx' inside the fusion matches the node
    Args:
        fusing_patterns: supported fusings
        node: node to decide if it can be a part of fusion
        idx: index of layer in the fusion
    Returns:
        fusing_patterns after filtering non-relevant fusions
    """
    valid_fusing_patterns = []
    for i,fusing_pattern in enumerate(fusing_patterns):
        if idx < len(fusing_pattern):
            if (type(fusing_pattern[idx]) == LayerFilterParams and fusing_pattern[idx].match(node)) or fusing_pattern[idx] == node.type:
                valid_fusing_patterns.append(fusing_pattern)

    # Return only valid patterns for this node
    return valid_fusing_patterns


def is_valid_fusion(fusing_patterns: List[List[Any]], nodes: List[BaseNode]) -> bool:
    """
    Check if the fusion is valid: exist in fusing_patterns
    Args:
        fusing_patterns: supported fusings
        nodes: nodes which are participating in fusion
    Returns:
        whether the fusion in valid
    """
    fusion_depth = len(nodes)
    if fusion_depth <= 1:
        return False
    for fusing_pattern in fusing_patterns:
        if fusion_depth != len(fusing_pattern):
            continue
        counter = 0
        for i,layer in enumerate(fusing_pattern):
            if (type(layer) == LayerFilterParams and layer.match(nodes[i])) or layer == nodes[i].type:
                counter += 1
        if counter == fusion_depth:
            return True
    return False


def disable_nodes_activation_quantization(nodes: List[BaseNode]):
    """
    Disable activation for non-quantization needed due to fusion
    Args:
        nodes: nodes to update their activation quantization
    """
    for node in nodes:
        for qc in node.candidates_quantization_cfg:
            qc.activation_quantization_cfg.enable_activation_quantization = False


def fusion(graph: Graph, tpc: TargetPlatformCapabilities) -> Graph:
    """
    Fusing defines a list of operators that should be combined and treated as a single operator,
    hence no quantization is applied between them when they appear in the graph.
    This function search and disable quantization for such patterns.
    Args:
        graph: Graph we apply the fusion on.
        tpc: TargetPlatformCapabilities object that describes the desired inference target platform (includes fusing patterns MCT should handle).
    Returns:
        Graph after applying fusion activation marking.
    """
    fusing_patterns = tpc.get_fusing_patterns()
    if len(fusing_patterns) == 0:
        return graph

    # Find max fusion
    max_layers_fusing = 0 if len(fusing_patterns) == 0 else max([len(fusing_pattern) for fusing_pattern in fusing_patterns])


    # -------------------------------- #
    # Fusion algorithm
    # -------------------------------- #
    fused_graph = copy.deepcopy(graph)

    # Travel along the graph to find layers for fusing
    nodes = fused_graph.get_topo_sorted_nodes()
    fused_nodes = []  # nodes that are participating in fusing
    for node in nodes:
        # Skip if already in fusing
        if node in fused_nodes:
            continue
        # Start fusing search
        fusing_nodes = [] # nodes that are candidates for participating in fusing
        patterns = copy.deepcopy(fusing_patterns)
        next_nodes = [node]
        for i in range(max_layers_fusing):
            patterns = filter_fusing_patterns(patterns, next_nodes[0], i)
            if len(patterns) == 0: # Give up if no more fusion pattern
                break
            fusing_nodes.append(next_nodes[0])
            next_nodes = fused_graph.get_next_nodes(fusing_nodes[-1])
            if len(next_nodes) != 1:  # Give up if node has more than one connection (not supported for fusion)
                break

        # New fusion: mark all nodes in the fusion except last one
        if is_valid_fusion(fusing_patterns, fusing_nodes):
            fused_nodes.extend(fusing_nodes)
            disable_nodes_activation_quantization(fusing_nodes[:-1])
            fused_graph.user_info.add_fusion(fusing_nodes)

    return fused_graph



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from model_compression_toolkit.core.common.constants import FOUND_TF, FOUND_TORCH
from model_compression_toolkit.core.tpc_models.default_tpc.v3.tp_model import get_tp_model, generate_tp_model, get_op_quantization_configs
if FOUND_TF:
    from model_compression_toolkit.core.tpc_models.default_tpc.v3.tpc_keras import get_keras_tpc as get_keras_tpc_latest
    from model_compression_toolkit.core.tpc_models.default_tpc.v3.tpc_keras import generate_keras_tpc
if FOUND_TORCH:
    from model_compression_toolkit.core.tpc_models.default_tpc.v3.tpc_pytorch import get_pytorch_tpc as get_pytorch_tpc_latest
    from model_compression_toolkit.core.tpc_models.default_tpc.v3.tpc_pytorch import generate_pytorch_tpc



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


# # Layer type constants:
PLACEHOLDER = 'placeholder'
OUTPUT = 'output'
CONSTANT = 'constant'
BUFFER = 'buffer'

# # Module operation type
CALL_FUNCTION = 'call_function'
CALL_METHOD = 'call_method'
GET_ATTR = 'get_attr'

# # Layers attributes constants:
KERNEL_SIZE = 'kernel_size'
PADDING = 'padding'
GROUPS = 'groups'
LAYER_NAME = 'name'
USE_BIAS = 'bias'
STRIDES = 'stride'
DILATIONS = 'dilation'
TENSOR_META = 'tensor_meta'
FILTERS = 'out_channels'
TYPE = 'type'
PAD = 'pad'
VALUE = 'value'
FUNCTIONAL_OP = 'functional_op'
OP_CALL_ARGS = 'op_call_args'
OP_CALL_KWARGS = 'op_call_kwargs'
INPUTS_AS_LIST = 'inputs_as_list'
INPLACE = 'inplace'
HARDTANH_MIN_VAL = 'min_val'
HARDTANH_MAX_VAL = 'max_val'

# # Layers variables names:
KERNEL = 'weight'
KERNEL_SIZE = 'kernel_size'
BIAS = 'bias'
GAMMA = 'weight'
BETA = 'bias'
IN_CHANNELS = 'in_channels'
MOVING_MEAN = 'running_mean'
MOVING_VARIANCE = 'running_var'
EPSILON = 'eps'
DIM = 'dim'
IN_CHANNELS = 'in_channels'
OUT_CHANNELS = 'out_channels'

# torch devices
CUDA = 'cuda'
CPU = 'cpu'

# ReLU bound constants
RELU_POT_BOUND = 8.0

# Supported TP models names for Pytorch:
DEFAULT_TP_MODEL = 'default'
TFLITE_TP_MODEL = 'tflite'
QNNPACK_TP_MODEL = 'qnnpack'

# MultiHeadAttention layer attributes:
EMBED_DIM = 'embed_dim'
NUM_HEADS = 'num_heads'
DROPOUT = 'dropout'
ADD_ZERO_ATTN = 'add_zero_attn'
KEY_DIM = "kdim"
VALUE_DIM = 'vdim'
BATCH_FIRST = 'batch_first'
OUT_PROJ_WEIGHT = 'out_proj.weight'
OUT_PROJ_BIAS = 'out_proj.bias'
V_PROJ_WEIGHT = 'v_proj_weight'
K_PROJ_WEIGHT = 'k_proj_weight'
Q_PROJ_WEIGHT = 'q_proj_weight'
IN_PROJ_WEIGHT = 'in_proj_weight'
IN_PROJ_BIAS = 'in_proj_bias'
BIAS_K = 'bias_k'
BIAS_V = 'bias_v'



# Copyright 2022 Sony Semiconductor Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch

def ste_round(x: torch.Tensor) -> torch.Tensor:
    """
    Calculate the rounded values of a tensor
    Args:
        x: input variable
    Returns:
        rounded value
    """
    return (torch.round(x) - x).detach() + x


def ste_clip(x: torch.Tensor, min_val=-1.0, max_val=1.0) -> torch.Tensor:
    """
    Clip a variable between fixed values such that min_val<=output<=max_val
    Args:
        x: input variable
        min_val: minimum value for clipping
        max_val: maximum value for clipping
    Returns:
        clipped variable
    """
    return (torch.clip(x, min=min_val, max=max_val) - x).detach() + x